from subvertpy import NODE_DIR

#from bzrlib import urlutils
from bzrlib.plugins.svn import errors as svn_errors
from bzrlib.plugins.svn.layout import  RepositoryLayout

def normalize_path(path):
    if path is None:
        return None
    path_comps = [p for p in path.split('/') if p]
    return '/' + '/'.join(path_comps)

def normalize_prefix_path(prefix):
    if prefix is None:
        return None
    _prefix = normalize_path(prefix) + '/'
    if _prefix != '//':
        return _prefix
    return '/'

def bzr_svn_path(path):
    return path[1:]

class SVNChangedPath(object):
    def __init__(self, path_property):
        (self.action, self.copyfrom_path, self.copyfrom_rev, self.kind) = path_property
        if self.copyfrom_path:
            self.copyfrom_path = '/' + self.copyfrom_path

    def __repr__(self):
        return ('action:' + self.action + ' kind:' + str(self.kind)
                 + ' from:(' + str(self.copyfrom_path) + ':' + str(self.copyfrom_rev) + ')\n')

class BranchInfo(object):
    pass

class AutoDetectCopyLayout(RepositoryLayout):
    #Override functions
    def __init__(self, _log=None, _prefix=None):
        if (_log is None):
            return
        if _prefix is None:
            _prefix = ""
        self.subdir = normalize_prefix_path(_prefix)
        self.log = _log
        print('AutoDetectCopyLayout with log', self.log, self.subdir)
        self.transport = _log._transport

        self.all_branches = {}

        self.last_revision = -1

        self.update_all()

        print('AutoDetectCopyLayout finished')
        #exit(0)

    def __repr__(self):
        return "%s" % (self.__class__.__name__)

    def __str__(self):
        return "auto"

#Implement interface
    def get_project_prefixes(self, project):
        return [project]

    def supports_tags(self):
        return True

    def get_tag_path(self, name, project=""):
        """Return the path at which the tag with specified name should be found.

        :param name: Name of the tag.
        :param project: Optional string identifier for the project the tag is
            for. Can include slashes.
        :return: Path of the tag.
        """
        return name

    def get_tag_name(self, path, project=""):
        """Determine the tag name from a tag path.

        :param path: Path inside the repository.
        """
        return path

    #def push_merged_revisions(self, project=""):TODO:

    def get_branch_path(self, name, project=""):
        return name

    def get_branch_name(self, path, project=""):
        """Determine the tag name from a tag path.

        :param path: Path inside the repository.
        """
        return path

    def parse(self, path):
        """Parse a path.

        :return: Tuple with type ('tag', 'branch'), project name, branch path and path
            inside the branch
        """
        print("parse %s:%s" % (path, self.revision))
        assert type(path) is str
        if path is None:
            raise svn_errors.NotSvnBranchPath(path)
        path_comps = [p for p in path.split('/') if p]
        if len(path_comps) == 0:
            raise svn_errors.NotSvnBranchPath(path)
        new_path = '/' + '/'.join(path_comps)
        subpath, branch_info = self.split_path(new_path, self.revision)
        if branch_info is None:
            raise svn_errors.NotSvnBranchPath(path)

        #If the current branch is not an real branch.
        if not branch_info[-1]:
            raise svn_errors.NotSvnBranchPath(path)

        path_type = self.get_path_type(branch_info[0], self.revision)
        return path_type, "", bzr_svn_path(branch_info[0]), subpath

    #def split_project_path(self, path, project):

    def is_branch(self, path, project=None):
        return self.get_path_type(path, self.revision) == 'branch'

    def is_tag(self, path, project=None):
        return self.get_path_type(path, self.revision) == 'tag'

    def is_branch_parent(self, path, project=None):
        #return self.is_branch(urlutils.join(path, "trunk"), project)
        return self.is_parent(path, project, 'branch')

    def is_tag_parent(self, path, project=None):
        return self.is_parent(path, project, 'tag')
    
    '''
    def is_branch_or_tag(self, path, project=None):
        return self.is_branch(path, project) or self.is_tag(path, project)

    def is_branch_or_tag_parent(self, path, project=None):
        return (self.is_branch_parent(path, project) or
                self.is_tag_parent(path, project))
    '''

    def get_branches(self, repository, revnum, project="", pb=None):
        """Retrieve a list of paths that refer to branches in a specific revision.

        :return: Iterator over tuples with (project, branch path)
        """
        print('get_branches', revnum)
        return self._get_paths(repository, revnum, project, pb, 'branch')

    def get_tags(self, repository, revnum, project="", pb=None):
        """Retrieve a list of paths that refer to tags in a specific revision.

        :return: Iterator over tuples with (project, branch path)
        """
        print('get_tags', revnum)
        return self._get_paths(repository, revnum, project, pb, 'tag')

#Local functions
    def _get_paths(self, repository, revnum, project, pb, path_type):
        for branch_path in self.all_branches:
            if self.get_path_type(branch_path, revnum) == path_type:
                yield "", bzr_svn_path(branch_path)

    def is_parent(self, path, project, path_type):
        result = False
        path_comps = [p for p in path.split('/') if p]
        for branch_path in self.all_branches:
            branch_comps = [p for p in branch_path.split('/') if p]
            if len(branch_comps) != len(path_comps) + 1:
                continue
            if branch_comps[:-1] != path_comps:
                continue
            if self.get_path_type(branch_path, self.revision) == path_type:
                result = True
                break
        print('is_parent', path, path_type, result)
        return result

    def get_path_type(self, branch_path, revision):
        normal_path = normalize_path(branch_path)
        branch_info = self.exist_branch(normal_path, revision)
        if not branch_info:
            return None
        if not branch_info[-1]:
            return None
        if branch_info[0].find("/tags/") >= 0:
            return 'tag'
        return 'branch'

    def get_paths(self, old_paths):
        paths = {}
        for p in sorted(old_paths.keys()):
            paths['/' + p] = SVNChangedPath(old_paths[p])
        return paths

    def update_all(self):
        latest_revnum = self.transport.get_latest_revnum()
        for (paths, i, _) in self.log.iter_changes(None, 1, latest_revnum):
            self.update(paths, i)

    def update(self, paths, revnum):
        """
        Given a revision object, determine changes to branches.
        """
        self.revision = revnum
        if (self.last_revision >= revnum):
            return
        self.last_revision = revnum
        paths = self.get_paths(paths)
        self.modified = {}
        sorted_paths = sorted(paths.keys())
        current_test_branches = frozenset(self.all_branches.keys())
        for p in sorted_paths:
            if paths[p].action != 'D' and  paths[p].action != 'R':
                continue
            np = p + '/'
            for branch_path in current_test_branches:
                if not self.exist_branch(branch_path, revnum):
                    continue
                if not (branch_path + '/').startswith(np):
                    continue
                self.remove_branch(branch_path, revnum)

        for p in sorted_paths:
            copy_from_path = paths[p].copyfrom_path
            if (copy_from_path is None):
                continue
            copy_from_path += '/'
            parent_rev = paths[p].copyfrom_rev
            for parent_path in current_test_branches:
                rn = parent_path + '/'
                if not rn.startswith(copy_from_path):
                    continue
                branch_info = self.exist_branch(parent_path, parent_rev)
                if not branch_info:
                    continue
                suffix = ''
                if (rn != copy_from_path):
                    suffix = parent_path[len(copy_from_path) - 1:]
                new_branch_path = p + suffix
                if self.can_be_branch(new_branch_path, revnum):
                    self.add_branch(new_branch_path, revnum, parent_path, parent_rev, branch_info[-1])

        for p in sorted_paths:
            #If p in an exist branch, then just continue
            if self.exist_branch(p, revnum):
                self.modified[p] = 1
                continue

            #If this path is deleted, then it's can not create new branches !
            if (paths[p].action == 'D'):
                continue

            #path should be an directory.
            if self.transport.check_path(p.lstrip('/'), revnum) != NODE_DIR:
                continue

            #Try to create an branch without care about copy from!
            self.create_branch(p, revnum, paths[p])

            branch_info = self.split_path(p, revnum)[1]
            if branch_info:
                self.modified[branch_info[0]] = 1
            else:
                print("WARNING:Path %s has copy from %s and not in an branch in revision:%s" % (p, paths[p].copyfrom_path, revnum))

    def create_branch(self, path, revision, path_info):
        test_branch_path = ''
        path_comps = [p for p in path.split('/') if p]
        while len(path_comps) != 0:
            test_branch_path += '/' + path_comps.pop(0)

            #If test_branch_path in an exist branch, then just continue
            if self.exist_branch(test_branch_path, revision):
                continue

            if (self.is_user_branch(test_branch_path, revision)):
                subpath = self.get_subpath(p, test_branch_path)
                parent_path, parent_rev = '/trunk', 0
                if subpath != '':
                    self.modified[test_branch_path] = 1
                print("Try add branch %s by using path %s on revision %s copy from %s!"
                      % (test_branch_path, p, revision, path_info.copyfrom_path))
                self.add_branch(test_branch_path, revision, parent_path, parent_rev,
                                self.is_real_branch(test_branch_path, revision))

    def can_be_branch(self, branch_path, revision):
        '''
        1.If USER specify the branch_path can be a branch, then it's a branch!
        2.Else if branch_path is contained in an solid-branch, then it's can not be an branch!
        3.Finally else, it's an branch! 
        '''

        #CHECK:branch_path should not be an branch at the current time!
        if self.exist_branch(branch_path, revision):
            raise Exception("The branch_path:%s revision:%s already be an branch, that's impossible!" % (branch_path, revision))

        #1.
        if self.is_user_branch(branch_path, revision):
            return True

        #2.
        for bp in self.all_branches:
            if branch_path.startswith(bp + '/') and self.is_solid_branch(bp, revision):
                print("%s in solid branch %s in revision:%s!" % (branch_path, bp, revision))
                return False
        return True

    def is_real_branch(self, branch_path, revision):
        """
        Judge if the branch is an real branch, that means if
        this branch_path will be convert to BZR. If it's not, that
        means this branch_path is only used to tracking the revision
        relation.
        """
        if branch_path.startswith(self.subdir):
            return True

        #TODO:If this branch is specified by the user, what's the property!
        return False

    def is_user_branch(self, branch_path, revision):
        #First, by default, 'trunk' is user specified branch!
        if branch_path.endswith('/trunk'):
            return True

        #Second, the folders under tags and branches should be treat as branch.
        path_comps = [p for p in branch_path.split('/') if p]
        if len(path_comps) >= 2 and (path_comps[-2] == "branches" or path_comps[-2] == "tags"):
            return True

        #TODO:Test if this branch_path is specified by the user!
        return False

    def is_solid_branch(self, branch_path, revision):
        '''We define an branch that can not contains other branches as solid-branch.
        Such as(/branches/a is an branch, then /branches/a/b can not be an branch,
        even if it's copied from existed branch(such as trunk)!).
        '''
        #First, we define trunk as solid-branch.
        if branch_path.endswith('/trunk'):
            return True

        #TODO:Test user specified solid-branch!
        return False

    def get_subpath(self, path, branch_path):
        if path == None or branch_path == None:
            return ''
        return path[len(branch_path) + 1:]

    def split_path(self, path, revision, include_close=False):
        if path == None:
            return '', None
        branch_info = None
        test_branch_path = ''
        path_comps = [p for p in path.split('/') if p]
        path = '/' + '/'.join(path_comps)
        while len(path_comps) != 0:
            test_branch_path += '/' + path_comps.pop(0)
            #Always using the deepest branch!
            branch_info = self.exist_branch(test_branch_path, revision, int(include_close)) or branch_info

        if (branch_info == None):
            return '', None
        subpath = self.get_subpath(path, branch_info[0])
        return subpath, branch_info

    def contains_branch(self, path, revnum):
        for branch_path, revlist in self.all_branches.iteritems():
            if not branch_path.startswith(path):
                continue
            if -1 in revlist:
                return True
        return False

    def add_branch(self, branch_path, open_revision, parent_path, parent_rev, is_real):
        branch_info = [parent_path, parent_rev, open_revision, is_real]
        print("Add branch %s with %s!" % (branch_path, branch_info))
        self.modified[branch_path] = 0
        if branch_path in self.all_branches:
            revlist = self.all_branches[branch_path]
            if -1 in revlist:
                revlist[-1][3] = branch_info[3]
            else:
                revlist[-1] = branch_info
        else:
            self.all_branches[branch_path] = {-1:branch_info}

    def remove_branch(self, branch_path, close_revision):
        self.modified[branch_path] = 0
        if branch_path in self.all_branches:
            if (-1 not in self.all_branches[branch_path]):
                raise Exception('Should contains key -1 in', self.all_branches[branch_path])
            branch_info = self.all_branches[branch_path].pop(-1)
            self.all_branches[branch_path][close_revision] = branch_info
        else:
            raise Exception('Remove an unexist branch', branch_path, 'in', self.all_branches)

    def exist_branch(self, branch_path, branch_rev, include_state=0):
        '''
        include_state = 0 means there is an internval [open,close) contains branch_rev
        include_state = 1 means there is an internval [open,close] contains branch_rev
        include_state = 2 means if there is an interval [open,close), then just return it.
            else return the interval that got the maximal close and also the close <= branch_rev
        '''
        if branch_path not in self.all_branches.iterkeys():
            return None
        revlist = self.all_branches[branch_path]
        result_revision = None
        if include_state == 1 and branch_rev in revlist:
            result_revision = branch_rev
        elif -1 in revlist and (branch_rev == -1 or revlist[-1][2] <= branch_rev):
            result_revision = -1
        else:
            for close_revision in revlist:
                open_revision = revlist[close_revision][2]
                if (open_revision <= branch_rev and branch_rev < close_revision):
                    result_revision = close_revision
                    break
        if result_revision is None and include_state == 2:
            result_revision = -2
            for close_revision in revlist:
                if close_revision <= branch_rev and close_revision > result_revision:
                    result_revision = close_revision
            if result_revision == -2:
                result_revision = None
        if result_revision is None:
            return None
        branch_info = list(revlist[result_revision])
        branch_info.insert(3, result_revision)
        return [branch_path, branch_rev] + branch_info

